<?php

namespace Drupal\Tests\aws_cloud\Functional\Ec2;

use Drupal\Tests\BrowserTestBase;
use Drupal\Component\Utility\Random;
use Drupal\Component\Serialization\Yaml;

abstract class AwsCloudTestCase extends BrowserTestBase {
  /**
   * Modules to enable.
   *
   * @var array
   */
  public static $modules = [
    'cloud', 'aws_cloud'
  ];

  /**
   * The profile to install as a basis for testing.
   *
   * @var string
   */
  protected $profile = 'minimal';

  protected $random;

  protected $cloud_context;


  /**
   * Set up test.
   */
  protected function setUp() {
    parent::setUp();

    if (!$this->random) {
      $this->random = new Random();
    }

    $this->createCloudContext();
    $this->initMockData();

    $perms = $this->getPermissions();
    $perms[] = 'view ' . $this->cloud_context;

    $web_user = $this->drupalCreateUser($perms);
    $this->drupalLogin($web_user);
  }

  /**
   * Get permissions of login user.
   *
   * @return permissions of login user.
   */
  abstract protected function getPermissions();

  /**
   * Get mock data.
   *
   * @return mock data.
   */
  protected function getMockData() {
    return [];
  }

  /**
   * Get mock data from configuration.
   *
   * @return mock data.
   */
  protected function getMockDataFromConfig() {
    $config = \Drupal::configFactory()->getEditable('aws_cloud.settings');
    return json_decode($config->get('aws_cloud_mock_data'), true);
  }

  /**
   * Update mock data in configuration.
   */
  protected function updateMockDataToConfig($mock_data) {
    $config = \Drupal::configFactory()->getEditable('aws_cloud.settings');
    $config->set('aws_cloud_mock_data', json_encode($mock_data))
           ->save();
  }

  private function createCloudContext() {
    $random = $this->random;
    $this->cloud_context = $random->name(8);

    $num = 1;
    $data = [
      'cloud_context'      => $this->cloud_context,
      'type'               => 'aws_ec2',
      'label'              => "Amazon EC2 US West ($num) - " . $random->name(8, TRUE),
    ];

    $cloud = entity_create('cloud_config', $data);
    $cloud->field_cloud_type->value = 'amazon_ec2';
    $cloud->field_description->value = "#$num: " . date('Y/m/d H:i:s - D M j G:i:s T Y') . $random->string(64, TRUE);
    $cloud->field_api_version->value = 'latest';
    $cloud->field_region->value = "us-west-$num";
    $cloud->field_api_endpoint_uri->value = "https://ec2.us-west-$num.amazonaws.com";
    $cloud->field_access_key->value = $random->name(20, TRUE);
    $cloud->field_secret_key->value = $random->name(40, TRUE);
    $cloud->field_user_id->value = $random->name(16, TRUE);
    $cloud->field_image_upload_url->value = "https://ec2.us-west-$num.amazonaws.com";
    $cloud->field_x_509_certificate->value = $random->string(255, TRUE);

    $cloud->save();
  }

  private function initMockData() {
    $mock_data = [];
    foreach ([__CLASS__, get_class($this)] as $class_name) {
      $content = $this->getMockDataFileContent($class_name, $this->getMockDataTemplateVars());
      if (!empty($content)) {
        $mock_data = array_merge($mock_data, Yaml::decode($content));
      }
    }

    $config = \Drupal::configFactory()->getEditable('aws_cloud.settings');
    $config->set('aws_cloud_test_mode', true)
           ->set('aws_cloud_mock_data', json_encode($mock_data))
           ->save();
  }

  /**
   * Get the content of mock data file.
   *
   * @return The content of mock data file.
   */
  protected function getMockDataFileContent($class_name, $vars, $suffix = '') {
    $path = drupal_realpath(drupal_get_path('module', 'aws_cloud')) . '/tests/mock_data';
    $pos = strpos($class_name, 'aws_cloud') + strlen('aws_cloud');
    $path .= str_replace('\\', '/', substr($class_name, $pos)) . $suffix . '.yml';

    if (!file_exists($path)) {
      return '';
    }

    $twig = \Drupal::service('twig');
    return $twig->renderInline(file_get_contents($path), $vars);
  }

  /**
   * Get variables in mock data template file.
   *
   * @return variables in mock data template file.
   */
  protected function getMockDataTemplateVars() {
    return [];
  }

  /**
   * Reload mock data in configuration.
   */
  protected function reloadMockData() {
    $mock_data = $this->getMockDataFromConfig();
    $file_content = $this->getMockDataFileContent(get_class($this), $this->getMockDataTemplateVars());
    if (!empty($file_content)) {
      $mock_data = array_merge($mock_data, Yaml::decode($file_content));
    }
    $this->updateMockDataToConfig($mock_data);
  }

}
